#!/usr/bin/env python

import time, socket, re, csv, argparse, signal, sys, os
import ssh

def debug (level, message) :
    if _debug >= level :
        sys.stderr.write("[debug] %s\n" % message.strip())
        sys.stderr.flush()

def log (message) :
    sys.stderr.write("[monitor] %s\n" % message.strip())
    sys.stderr.flush()

from subprocess import Popen, PIPE

class Monitor (object) :
    def run (self, timeout, logpath, halt, normalise, pid, pidfile) :
        maxmem, maxcpu, now, _now = 0, 0, 0, 0
        try :
            debug(1, "Monitor.run: getting first values")
            now, mem, cpu = _now, _mem, _cpu = self.top()
            debug(1, "Monitor.run: stopping parent")
            os.kill(pid, signal.SIGTERM)
            with open(pidfile, "w") as pf :
                pf.write("%s\n" % os.getpid())
            status = "timeout"
            debug(1, "Monitor.run: opening logfile")
            with open(logpath, "wb") as logfile :
                log = csv.writer(logfile)
                start = time.time()
                debug(1, "Monitor.run: entering mainloop (please wait...)")
                while True :
                    now, mem, cpu = self.top()
                    if normalise :
                        now -= _now
                        mem -= _mem
                        cpu -= _cpu
                    maxmem = max(maxmem, mem)
                    maxcpu = max(maxcpu, cpu)
                    log.writerow([now, mem, cpu])
                    logfile.flush()
                    if timeout and time.time() - start > timeout:
                        _now = now
                        break
        except KeyboardInterrupt :
            debug(1, "Monitor.run: interrupted")
            status = "normal"
        if halt :
            debug(1, "Monitor.run: halting VM")
            self.halt()
        return maxmem, maxcpu, now - _now, status

def regexp (expr) :
    return re.compile(expr.format(cpu_id="(?P<cpu_id>[0-9.]+)",
                                  mem_total="(?P<mem_total>[0-9]+[kmg]?)",
                                  mem_used="(?P<mem_used>[0-9]+[kmg]?)",
                                  mem_unit="(?P<mem_unit>[kmg])"),
                      re.I)

class LinuxMonitor (Monitor) :
    _mem = {"k" : 1024, "m" : 1024**2, "g" : 1024**3}
    @classmethod
    def parse_mem (self, text) :
        return int(text[:-1]) * self._mem[text[-1].lower()]
    _top = regexp(r"cpu\(s\):.*,\s*{cpu_id}%id,.*\nmem:\s*{mem_total}\s*total,\s*{mem_used}\s*used,")
    _top = regexp(r"cpu\(s\):.*,\s*{cpu_id}\s*id,.*\n{mem_unit}\S*\s*mem:\s*{mem_total}\s*total,\s*{mem_used}\s*used,")
    @classmethod
    def parse (self, text) :
        debug(3, "LinuxMonitor.parse: matching %r" % text)
        data = self._top.search(text).groupdict()
        if "mem_unit" in data :
            if data["mem_total"][-1] in "0123456789":
                data["mem_total"] += data["mem_unit"]
            if data["mem_used"][-1] in "0123456789":
                data["mem_used"] += data["mem_unit"]
        total, used = [self.parse_mem(x)
                       for x in [data["mem_total"], data["mem_used"]]]
        debug(4, "got: total=%s, used=%s" % (total, used))
        return (100.0*used)/total, 100 - float(data["cpu_id"])
    def __init__ (self, hostname, port, username, keypath) :
        debug(1, "LinuxMonitor: socket.socket()")
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        debug(1, "LinuxMonitor: socket.connect((%s, %s))" % (hostname, port))
        sock.connect((hostname, port))
        debug(1, "LinuxMonitor: wrapping socket to SSH")
        self.ssh = ssh.Transport(sock)
        self.ssh.start_client()
        debug(2, "LinuxMonitor: load private key")
        key = ssh.RSAKey.from_private_key_file(keypath)
        debug(1, "LinuxMonitor: login as %s" % username)
        self.ssh.auth_publickey(username, key)
        debug(1, "LinuxMonitor: dummy top")
        self.top()
        debug(1, "LinuxMonitor: init done")
    def top (self) :
        debug(2, "LinuxMonitor.top: ssh.open_session()")
        chan = self.ssh.open_session()
        stdin = chan.makefile("w")
        stdout = chan.makefile("r")
        debug(2, "LinuxMonitor.top: ssh.exec_command('top -b -n 1|head -n 5')")
        chan.exec_command("top -b -n 1|head -n 5")
        output = stdout.read().rstrip()
        debug(3, "got: %r" % output)
        mem, cpu = self.parse(output)
        debug(3, "parsed to: mem=%s, cpu=%s" % (mem, cpu))
        chan.close()
        return time.time(), mem, cpu
    def halt (self) :
        try :
            debug(1, "LinuxMonitor.halt: ssh.open_session()")
            chan = self.ssh.open_session()
            debug(1, "LinuxMonitor.halt: ssh.exec_command('halt')")
            chan.exec_command("halt")
        except Exception as err :
            debug(1, "raised %s: %s" % (err.__class__.__name, err))
            pass

class WindowsMonitor (Monitor) :
    def __init__ (self, hostname, port, username, keypath, nsclient, checknt) :
        self.hostname = hostname
        self.port = port
        self.username = username
        self.keypath = keypath
        self.nsclient = nsclient
        self._checknt_mem = "%s -H localhost -p %s -v MEMUSE" % (checknt, nsclient)
        self._checknt_cpu = "%s -H localhost -p %s -v CPULOAD -l 1,100,100" % (checknt, nsclient)
        self.top()
    def halt (self) :
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.hostname, self.port))
        self.ssh = ssh.Transport(sock)
        self.ssh.start_client()
        key = ssh.RSAKey.from_private_key_file(self.keypath)
        self.ssh.auth_publickey(self.username, key)
        chan = self.ssh.open_session()
        chan.exec_command("shutdown /p /f")
    def _shell (self, command) :
        return Popen(command.split(), stdout=PIPE).communicate()[0]
    def top (self) :
        start = time.time()
        mem = self._check_mem()
        cpu = self._check_cpu()
        stop = time.time()
        if stop - start < 1 :
            time.sleep(1 - stop + start)
        return stop, mem, cpu
    _mem = re.compile("used:[^()]*\\((?P<mem>[0-9.]+)%\\)")
    def _check_mem (self) :
        text =  self._shell(self._checknt_mem)
        data = self._mem.search(text).groupdict()
        return float(data["mem"])
    _cpu = re.compile("CPU Load (?P<cpu>[0-9.]+)%")
    def _check_cpu (self) :
        text =  self._shell(self._checknt_cpu)
        data = self._cpu.search(text).groupdict()
        return float(data["cpu"])

if __name__ == "__main__" :
    global _debug
    parser = argparse.ArgumentParser(description="Monitor a VM for MCC contest.")
    parser.add_argument("--debug", dest="debug", metavar="LEVEL",
                        type=int, default=0,
                        help="print debug messages")
    parser.add_argument("--timeout", metavar="NUM", type=int,
                        dest="timeout", default=0,
                        help="stop logging after NUM seconds, 0 to disable (default)")
    parser.add_argument("--tries", metavar="NUM", type=int,
                        dest="tries", default=0,
                        help="tries to connect NUM times, 0 is forever (default)")
    parser.add_argument("--pidfile", metavar="PATH", type=str,
                        dest="pidfile", default="monitor.pid",
                        help="save monitor's PID in PATH (default 'monitor.pid')")
    parser.add_argument("--log", metavar="PATH", type=str,
                        dest="log", default="monitor.log",
                        help="write logs to PATH in CSV format (date,%%MEM,%%CPU)")
    parser.add_argument("--top-regexp", metavar="RE", type=str,
                        dest="top", default="",
                        help="regexp to match top output")
    parser.add_argument("--summary", metavar="PATH", type=str,
                        dest="summary", default="summary.log",
                        help="write one-line summary to PATH (default 'summary.log')")
    parser.add_argument("--priv", metavar="PRIV", type=str,
                        dest="priv", default="mcc.priv",
                        help="load private key from file PRIV (default 'mcc.priv')")
    parser.add_argument("--keep-run", action="store_false",
                        dest="halt", default=True,
                        help="keep VM running after logging has finished")
    parser.add_argument("--normalise", action="store_true",
                        dest="normalise", default=False,
                        help="normalise results wrt initial state")
    parser.add_argument("--nsclient", metavar="PORT", type=int,
                        dest="nsclient", default=0,
                        help="log Windows/NSClient++ on PORT (default Linux)")
    parser.add_argument("--check-nt", metavar="PATH", type=str,
                        dest="checknt", default="/usr/lib/nagios/plugins/check_nt",
                        help="check_nt executable (default '/usr/lib/nagios/plugins/check_nt')")
    parser.add_argument("login", metavar="USER@HOST:PORT", type=str,
                        help="parameters for SSH connection")
    parser.add_argument("format", metavar="FORMAT", type=str,
                        help="format for one-line summary (use '{MEM}', '{CPU}', '{TIME}' and '{STATUS}')")

    try :
        args = parser.parse_args()
    except :
        log("invalid command line")
        parser.print_help(sys.stderr)
        sys.exit(1)

    _debug = args.debug
    LinuxMonitor._top = regexp(open(args.top).read().strip())

    match = re.match("^(?P<username>[^@]+)@(?P<hostname>[^:]+):(?P<port>[0-9]+)$",
                     args.login)
    if match is None :
        log("invalid command line")
        parser.print_help(sys.stderr)
        sys.exit(1)

    login = match.groupdict()

    # detach process
    debug(1, "parent: spawning monitor")
    parent = os.getpid()
    if os.fork() :
        # parent waits for signal
        def stop (signum, frame) :
            if signum == signal.SIGTERM :
                sys.exit(0)
            elif signum == signal.SIGUSR1 :
                log("could not connect to VM")
                sys.exit(2)
            elif signum == signal.SIGUSR2 :
                log("monitoring process has crashed")
                sys.exit(3)
        signal.signal(signal.SIGTERM, stop)
        signal.signal(signal.SIGUSR1, stop)
        signal.signal(signal.SIGUSR2, stop)
        while True :
            time.sleep(args.timeout or 10)
    # double fork
    if os.fork() :
        sys.exit()

    def term (signum, frame) :
        raise KeyboardInterrupt
    signal.signal(signal.SIGTERM, term)

    top, tries = None, args.tries
    while args.tries == 0 or tries > 0 :
        debug(1, "trying to connect to the WM")
        try :
            if args.nsclient :
                top = WindowsMonitor(username=login["username"],
                                     hostname=login["hostname"],
                                     port=int(login["port"]),
                                     keypath=args.priv,
                                     nsclient=args.nsclient,
                                     checknt=args.checknt)
            else :
                top = LinuxMonitor(username=login["username"],
                                   hostname=login["hostname"],
                                   port=int(login["port"]),
                                   keypath=args.priv)
            break
        except Exception as err :
            debug(1, "raised %s: %s" % (err.__class__.__name__, err))
            time.sleep(1)
            if args.tries :
                tries -= 1

    if top is None :
        debug(1, "top is None")
        os.kill(parent, signal.SIGUSR1)
        sys.exit(2)

    try :
        debug(1, "running monitor")
        mem, cpu, sec, status = top.run(timeout=args.timeout,
                                        logpath=args.log,
                                        halt=args.halt,
                                        normalise=args.normalise,
                                        pid=parent,
                                        pidfile=args.pidfile)
    except KeyboardInterrupt :
        debug(1, "interrupted")
        mem, cpu, sec, status = 0, 0, 0, "skipped"
    except (socket.error, ssh.SSHException) as err :
        debug(1, "network error %s: %s" % (err.__class__.__name__, err))
        try :
            os.kill(parent, signal.SIGUSR1)
        except :
            pass
        sys.exit(2)
    except Exception as err :
        log("raised %s: %s" % (err.__class__.__name__, err))
        try :
            os.kill(parent, signal.SIGUSR2)
        except :
            pass
        sys.exit(3)

    def term (signum, frame) :
        pass
    signal.signal(signal.SIGTERM, term)

    with open(args.summary, "ab") as logfile :
        logfile.write(args.format.strip().format(MEM=mem, CPU=cpu,
                                                 TIME=sec, STATUS=status)
                      + "\n")

    try :
        os.kill(parent, signal.SIGUSR1)
    except :
        pass
